import csv
import math
import numpy as np
from operator import itemgetter
import pickle
import random
from scipy.spatial.distance import cityblock

#============================== User defined parameters ==============================

# Testing the algorithm for various value of 'K' and 'Latency'
K_lst = [5,10,50,100]
latency_lst = [100,200,500,1000,2000,2500,5000]  # in seconds, latency >= K

#=====================================================================================

# Accepts a new record and stores it for further processing
def process_new_record(row):
    global data
    data.append(float(row[1]))

# Processes the records and computes generalized values
def return_publishable_records(time_step, end_flag):
    global buffered_vals, data_out
    pub_records = []
    records_sorted = sorted(buffered_vals, key = itemgetter(0))

    # If the data stream has reached its end then generalize all records in the buffer
    if end_flag == True:
        '''If there are fewer than K records then compute the mean and look for the value
        closest to mean in the published records and assign all records that value'''
        if len(records_sorted) < K:
            current_vals = records_sorted
            mean_val = np.mean([item[0] for item in current_vals])
            data_out_val = [item[3] for item in data_out]

            current_vals_out = [min(data_out_val, key=lambda x:abs(x-mean_val))]*len(current_vals)

            for j in range(len(current_vals)):
                    pub_records.append([time_step-current_vals[j][1], time_step, current_vals[j][0], current_vals_out[j]])

        # Otherwise process all records, K at a time making sure last batch of records is at least K in size
        else:
            q = len(records_sorted) // K
            i = 0
            while i <= q-1:
                if i == q-1:
                    x = len(records_sorted)+1
                else:
                    x = (i+1)*K

                current_vals = records_sorted[i*K:x]
                current_vals_out = [np.mean([item[0] for item in current_vals])]*(x-i*K)

                for j in range(len(current_vals)):
                    pub_records.append([time_step-current_vals[j][1], time_step, current_vals[j][0], current_vals_out[j]])

                i += 1

    # The first value in buffer is closest to its latency, check if it has expired
    elif buffered_vals[0][1] == latency:
        # If the records in buffer are fewer than K then assign a value from published values as before
        if len(records_sorted) <= K:
            current_vals = records_sorted
            mean_val = np.mean([item[0] for item in current_vals])
            data_out_val = [item[3] for item in data_out]

            current_vals_out = [min(data_out_val, key=lambda x:abs(x-mean_val))]*len(current_vals)

        # If more than K records are present then process only the record that is expiring
        else:
            record_idx = records_sorted.index(buffered_vals[0])

            bottom_idx = max(record_idx-K+1, 0)
            top_idx = min(len(records_sorted)-1, record_idx+K-1)

            idx = bottom_idx
            current_distortion = math.inf

            # Sort the records and look for the best grouping of the record that is expiring to minimize distortion
            while idx+K-1 <= top_idx:
                new_vals = records_sorted[idx:idx+K]
                new_vals_out = [np.mean([item[0] for item in new_vals])]*K
                new_distortion = cityblock([item[0] for item in new_vals], new_vals_out)

                if new_distortion < current_distortion:
                    current_vals = new_vals
                    current_distortion = new_distortion
                    current_vals_out = new_vals_out

                idx += 1

        # Publishable records that are returned with time in, time out, original value, generalized value
        for i in range(len(current_vals)):
            pub_records.append([time_step-current_vals[i][1], time_step, current_vals[i][0], current_vals_out[i]])

        buffered_vals = [item for item in buffered_vals if item not in current_vals]

    return pub_records


# Load the dataset
data = []
with open('data.csv', 'r') as f:
    reader = csv.reader(f)
    for row in reader:
        process_new_record(row)

# Compute the generalized values for various value of 'K' and 'Latency'
for K in K_lst:
    for latency in latency_lst:
        data_out = []
        buffered_vals = []

        '''Each iteration is one time-step'''
        end_flag = False
        for time_step in range(len(data)):
            # If the buffer is not empty then increase the time spent by all data points
            if len(buffered_vals) > 0:
                buffered_vals = [[item[0], item[1]+1] for item in buffered_vals]

            # Set flag True if the data stream has ended
            if time_step == len(data) - 1:
                end_flag = True

            # Add new value to buffer with time spend set to 0
            buffered_vals.append([data[time_step], 0])

            # Get publishable records
            pub_records = return_publishable_records(time_step, end_flag)

            if len(pub_records) > 0:
                data_out.extend(pub_records)

        print('\n=========== K:', K, '=========== Latency:', latency, ' ================')
        pickle.dump(data_out, open('output/data_out_' + str(K) + '_' + str(latency) + '.pkl', "wb"))
