package com.privitar;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.List;

import com.privitar.InputRecord;
import com.privitar.OutputRecord;
import com.privitar.StreamingKFilter;
import com.privitar.measures.RmseDistortion;

public class MyFilter implements StreamingKFilter{
	
	private final int K;
	private List<InputRecord > inputBuffer;
	private List<OutputRecord> outputBuffer;
	private double minValue;
	private double maxValue;
	private int currentTime;
	private final Integer blockSize;
	private final Integer maxWaitTime;
	
	
	public MyFilter() {
        this.K           = 5;         // k-anonymity parameter
    	this.blockSize   = 30;        // maximum number of input records before starting to compute the output records
    	this.maxWaitTime = 100000000; // maximum wait time before starting to compute the output records
        
    	assert (this.blockSize >= this.K);
    	
    	// initialize input and output buffers
        this.inputBuffer  = new ArrayList<InputRecord>();
        this.outputBuffer = new ArrayList<OutputRecord>();
        
        // initialize min and max values 
        this.minValue = Double.MAX_VALUE;
        this.maxValue = Double.MIN_VALUE;
    }
	
	/*
	 * Generate output records when enough input records are received
	 */
	private void generateOutputRecords() {
		int size;
		double distortion, anonymizedValue, distortion_min = Double.MAX_VALUE;
		
		this.inputBuffer.sort(Comparator.comparing(InputRecord::getRawValue));
		
		while (this.inputBuffer.isEmpty() == false) {
			
			// initialize the variables used in the search
			distortion_min = Double.MAX_VALUE;
			size = this.K -1;
			
			do {
				// process 'size' number of records
				size += 1;
				
				// process 'size' number of objects, if the remaining are less than this.K then process them all
				if (size > (this.inputBuffer.size() - this.K) )
					size = this.inputBuffer.size();
				
				// compute the distortion and the anonymized value on the records
				anonymizedValue = getAnonValue (this.inputBuffer.subList(0, size));
				distortion      = getDistortion(this.inputBuffer.subList(0, size), anonymizedValue);
				
				// if the distortion is decreasing when adding records then keep adding elements
				// otherwise output them 
				if (distortion_min > distortion)
					distortion_min = distortion;
				else
					break;
				
			} 
			while(size < this.inputBuffer.size());
			
			// add the records to the output buffer
			addOutputRecords(this.inputBuffer.subList(0, size), anonymizedValue);
			
			// continue to process the other records in the input buffer 
			this.inputBuffer = this.inputBuffer.subList(size, this.inputBuffer.size());	
		}
		
		// Clear the input buffer
    	this.inputBuffer = new ArrayList<InputRecord>();
	}	
	
	/*
	 * Calculate the anonymized value of the input records
	 */
	private double getAnonValue(List<InputRecord> records) {
		double average = 0;
		for (InputRecord in : records)
	        average += in.getRawValue();
		return average / records.size();
	}
	
	/*
	 *  Add input records to the records ready to be published
	 */
	private void addOutputRecords(List<InputRecord> records, double anonValue) {
		for (InputRecord in: records)
			this.outputBuffer.add(new OutputRecord(in, this.currentTime, anonValue));
	}
	
	/*
	 * Calculate the error metrics produced by the input records with the anonymized value specified
	 * The error metric is taken from the test suite 
	 */
	private double getDistortion(List<InputRecord> records, double anonValue) {
		double minValue, maxValue;
		List<OutputRecord> outTmp = new ArrayList<OutputRecord>();
		
		minValue = this.inputBuffer.get(0).getRawValue();
		maxValue = this.inputBuffer.get(this.inputBuffer.size()-1).getRawValue();
		if (this.minValue > minValue)
			this.minValue = minValue;
		if (this.maxValue < maxValue)
			this.maxValue = maxValue;
				
		for (InputRecord input: records)
			outTmp.add(new OutputRecord(input, this.currentTime, anonValue));
	
		return RmseDistortion.measureDataDistortion(outTmp, this.maxValue - this.minValue);
	}
    
	@Override
	public void processNewRecord(InputRecord input) {
		Integer firstElementTime, lastElementTime, elapsedTime;
		
		this.inputBuffer.add(input);
		
		// calculating elapsed time
		firstElementTime = this.inputBuffer.get(0).getTime();
		lastElementTime  = this.inputBuffer.get(this.inputBuffer.size()-1).getTime();
		elapsedTime      = firstElementTime - lastElementTime;
		
		// update the current time
		this.currentTime = input.getTime();
		
		// if enough records are received then process the block
		if (this.inputBuffer.size() >= this.blockSize)
			generateOutputRecords();
		
		// if enough time is elapsed then process the block
		if (elapsedTime >= this.maxWaitTime && this.inputBuffer.size() >= this.K) {
			generateOutputRecords();
		}
	}

	@Override
	public Collection<OutputRecord> returnPublishableRecords() {
		Collection<OutputRecord > outputs = this.outputBuffer;
		
		this.outputBuffer = new ArrayList<OutputRecord >();
		return outputs;
	}
}
